/*
 * umoci: Umoci Modifies Open Containers' Images
 * Copyright (C) 2016-2020 SUSE LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package umoci

import (
	"fmt"
)

// These are populated during "make" using -ldflags "-X ...".
// TODO: Switch to embedding the version here so that FullVersion() makes sense
//       even when umoci is used as a library.
var (
	version   = "unknown"
	gitCommit = ""
)

// FullVersion returns a fully-qualified version string if one is available.
// NOTE: This function will return "unknown" if umoci is being used as a "go
//       get" dependency or binary.
func FullVersion() string {
	v := "unknown"
	if version != "" {
		v = version
	}
	if gitCommit != "" {
		v = fmt.Sprintf("%s~git%s", v, gitCommit)
	}
	return v
}
